/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2006-2013 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 */
#ifndef TOPPERS_SOCFPGA_H
#define TOPPERS_SOCFPGA_H

#include <sil.h>

/*
 * CPU clocks
 */
#define SOCFPGA_CORE_CLK           800U
#define SOCFPGA_PERIPHCLK          (SOCFPGA_CORE_CLK/4U)
#define MPCORE_GTC_CLOCK_FREQ_MHZ  SOCFPGA_PERIPHCLK

/*
 * Tmer preescaler and load value to achieve 1ms per tick.
 * Note that the preescaler value must be representable with 8 bits;
 * and the load value must be a 32bit value.
 *
 * Private timers and watchdogs are clocked with PERIPHCLK which is 1/2 of
 * the CPU frequency. The formula is:
 *
 *    period = (preescaler + 1)(load + 1) / PERIPHCLK
 *      - period    = 1ms
 *      - periphclk = 200MHz
 *
 * For example:
 *      - preescaler = 0
 *      - load       = 200000
 */
#define MPCORE_WDT_PS_1MS    0U                          /* preescaler for private watchdog */
#define MPCORE_WDT_LOAD_1MS  (SOCFPGA_PERIPHCLK * 1000U) /* load value for private watchdog */

#define MPCORE_TM_PS_1MS     0U                          /* preescaler for private timer */
#define MPCORE_TM_LOAD_1MS   (SOCFPGA_PERIPHCLK * 1000U) /* load value for private timer */

/*
 * Memory base address and size
 */
#define DDR_ADDR 0x00000000
#define DDR_SIZE 0x40000000 /* 1000MB */

#define OSRAM_ADDR 0xFFFF0000

/*
 *  MPCore Private Memory Region Base Address
 */
#define MPCORE_PMR_BASE  0xfffec000

/*
 *  Number of interrupts supported by the GICv1.0 in this board. Note
 *  that these values could be obtained dynamically from the corresponding
 *  GIC register.
 */
#define DIC_TMIN_INTNO    0U
#define DIC_TMAX_INTNO   211U // maximum interrupt ID
#define DIC_TNUM_INT     212U // maximum number of interrupts

/*
 *
 */
#define H2F_OFFSET    0xC0000000
#define LWH2F_OFFSET  0xff200000

/*
 *  RSTMGR
 */
#define RSTMGR_BASE 0xFFD05000
#define RSTMGR_MPUMODRST  (RSTMGR_BASE + 0x10)
#define RSTMGR_PERMODRST  (RSTMGR_BASE + 0x14)
#define RSTMGR_BRGMODRST  (RSTMGR_BASE + 0x1C)

#define RSTMGR_MPUMODRST_CPU1  0x02

#define RSTMGR_PERMODRST_L4WD0 0x40

#define RSTMGR_BRGMODRST_H2F   0x01
#define RSTMGR_BRGMODRST_LWH2F 0x02 
#define RSTMGR_BRGMODRST_F2H   0x04

/*
 *  SYSMGR
 */
#define SYSMGR_BASE 0xFFD08000
#define SYSMGR_CPU1STARTADDR (SYSMGR_BASE + 0xC4)

/*
 *  FPGAMGR
 */
#define FPGAMGR_BASE 0xFF706000
#define FPGAMGR_CTRL (SYSMGR_BASE + 0x04)
  
/*
 *  L3REGS
 */
#define L3REGS_BASE 0xFF800000
#define L3REGS_REMAP (L3REGS_BASE + 0x00)

#define L3REGS_REMAP_MPUZERO     0x01
#define L3REGS_REMAP_NONMPUZERO  0x02
#define L3REGS_REMAP_HPS2FPGA    0x08
#define L3REGS_REMAP_LWHP2FPGA   0x10

/*
 *  MPUL2
 */
#define MPUL2_BASE 0xFFFEF000

#define MPUL2_ADRFILTER (MPUL2_BASE  + 0xC00)

/*
 *  UART
 */
#define UART0_BASE  0xFFC02000
#define UART1_BASE  0xFFC03000

#define UART0_LCR_VAL 0
#define UART0_DLM_VAL 0
#define UART0_DLL_VAL 0

#define UART1_LCR_VAL 0
#define UART1_DLM_VAL 0
#define UART1_DLL_VAL 0

/*
 *  Interrupt Number
 */
#define DIC_IRQNO_UART0 194
#define DIC_IRQNO_UART1 195


#ifndef TOPPERS_MACRO_ONLY

/*
 *  Trace log definitions (TODO: not tested)
 *
 *  TRACE_GET_TIM returns microseconds and uses the lower part of the
 *  global timer counter (32bit). This means that only around 12.89 seconds
 *  (0xFFFFFFFF/333 us) of trace can be taken until the counter wraps around.
 *
 *  Note: For traces that take more time, comment the TRACE_HW_INIT and
 *  TRACE_GET_TIM. As defined in trace_config.h, get_my_current_time()
 *  will be used instead. This allows taking a very long trace log (up to
 *  49,7 days) but the resolution will be in the milliseconds order.
 */
#ifndef MPCORE_GTC_COUNT_L
#define MPCORE_GTC_COUNT_L (MPCORE_PMR_BASE + 0x0200)
#endif  /* MPCORE_GTC_COUNT_L */

#define TRACE_HW_INIT() x_init_pcc()
#define TRACE_GET_TIM() (*((volatile uint32_t *)(MPCORE_GTC_COUNT_L)) / (MPCORE_GTC_CLOCK_FREQ_MHZ))

#endif /* TOPPERS_MACRO_ONLY */
#endif /* TOPPERS_SOCFPGA_H */
