/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2007-2013 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: target_config.c 1019 2013-07-11 13:29:21Z ertl-honda $  
 */

#include "kernel_impl.h"
#include <sil.h>
#include "ns16550.h"
#include "pl310.h"

/*
 *  entry point (start.S)
 */
extern void start(void);

/*
 *  PPRC2 boot code(target_support.S)
 */
extern uint8_t prc2_boot;
extern uint8_t prc2_boot_end;

static void
copy_prc2_boot(void){
	int cnt;
	volatile uint8_t *dst_addr;

	dst_addr = (uint8_t*)0x00000000; /* DDR */
	for (cnt = 0; cnt < (&prc2_boot_end - &prc2_boot); cnt++) {
		*(dst_addr + cnt) = *(volatile uint8_t *)(&prc2_boot + cnt);
	}

	__asm__ volatile("dsb");
	__asm__ volatile("isb");
}


/*
 *  Master processor initialization before str_ker().
 *  (E.g.: initialize the interrupt controller or the SCU.)
 */
void
target_mprc_initialize(void)
{
	chip_mprc_initialize();

#ifndef TOPPERS_SAFEG_NONSECURE
	/* map DDR to 0x00000000 - 0xffff0000 (also for uboot)*/
	sil_wrw_mem((uint32_t*)MPUL2_ADRFILTER, 0x01);
#endif  /* TOPPERS_SAFEG_NONSECURE */

#if TNUM_PRCID >= 2

#if !defined(TOPPERS_SAFEG_SECURE) & !defined(TOPPERS_SAFEG_NONSECURE)
	/*
	 * SafeG 未使用時
	 */
	/* Core2 boot code to DDR */
	copy_prc2_boot();
	/* Set Core2 boot address */
	sil_wrw_mem((uint32_t*)SYSMGR_CPU1STARTADDR, (uint32_t)start);
	/* Wake up processor 2 */
	sil_wrw_mem((uint32_t*)RSTMGR_MPUMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_MPUMODRST) & ~RSTMGR_MPUMODRST_CPU1);
#endif /* !defined(TOPPERS_SAFEG_SECURE) & !defined(TOPPERS_SAFEG_NONSECURE) */

#ifdef TOPPERS_SAFEG_NONSECURE
	/*
	 *  SafeG 使用時(NonSecure側) 
	 *  プロセッサ2はSecure側がOSRAM上にコピーした first_switch_to_nt
	 *  で SYSMGR_CPU1STARTADDR が設定されるのを待っているため，
	 *  起動させるために書き込む．
	 */
	sil_wrw_mem((uint32_t*)SYSMGR_CPU1STARTADDR, (uint32_t)start);
#endif /* TOPPERS_SAFEG_NONSECURE */

#endif /* TNUM_PRCID >= 2 */
}

void
target_fput_log(char_t c)
{
	if (c == '\n') {
		sio_pol_putc('\r', 1);
	}
		sio_pol_putc(c, 1);
}

/*
 * MMU initialization (called from the chip code)
 *
 * We create a simpe 1-to-1 (physical address = virtual address) memory
 * mapping and enable caching for the DDR memory.
 */
void
target_mmu_init(void)
{
	MEMORY_ATTRIBUTE m_attribute;

	/*
	 * Basic initialization of the whole virtual memory space to
	 * non-cachable, non-bufferable, strongly-ordered.
	 */
	m_attribute.pa   = 0x00000000;
	m_attribute.va   = m_attribute.pa;
	m_attribute.size = 0x80000000; /* 8GB */
	m_attribute.ns   = 0;          /* 0=Secure */
	m_attribute.s    = 1;          /* 1=Shared */
	m_attribute.ap   = 3;          /* Full access */
	m_attribute.tex  = 0;          /* Strongly Ordered */
	m_attribute.c    = 0;
	m_attribute.b    = 0;

	mmu_map_memory (&m_attribute);

	m_attribute.pa   = 0x80000000; /* 2GB */
	m_attribute.va   = m_attribute.pa;

	mmu_map_memory (&m_attribute);

	/*
	 *  Enable caching for the DDR
	 */
	m_attribute.pa   = DDR_ADDR;
	m_attribute.va   = m_attribute.pa;
	m_attribute.size = DDR_SIZE;
	m_attribute.ns   = 0;      /* NonSecure    */
	m_attribute.s    = 1;      /* 1=Shared */
	m_attribute.ap   = 3;      /* Full access */
	/* Outer and Inner Write-Back, Write-Allocate */
	m_attribute.tex  = 1;             
	m_attribute.c    = 1;             
	m_attribute.b    = 1;             

	mmu_map_memory(&m_attribute);

#ifdef TOPPERS_SAFEG_SECURE
	/*
	 *  Enable caching for the DDR
	 */
	m_attribute.pa   = DDR_ADDR;
	m_attribute.va   = m_attribute.pa;
	m_attribute.size = 0x3F000000;
	m_attribute.ns   = 1;        /* NonSecure    */
	m_attribute.s    = 1;        /* 1=Shared */
	m_attribute.ap   = 3;        /* Full access */

    /* non-cache   ok */
//	m_attribute.tex  = 1;
//	m_attribute.c    = 0;
//	m_attribute.b    = 0;

    /* Outer and Inner Write-Back, Write-Allocate */
//	m_attribute.tex  = 1;
//	m_attribute.c    = 1;
//	m_attribute.b    = 1;

	/*
	 * この設定でないと，L2の初期化のためにLinuxがidを取得した際にSafeGから
	 * Linux変数に書き込んだ値が反映されない.
	 * L2を有効した後に有効かは要検討
	 */
	
	/* スタック上の変数の場合はこちらで動作する */
	/* Onter and Inner Write-Back; no Write-Allocate ng */ 
	m_attribute.tex  = 0;
	m_attribute.c    = 1;
	m_attribute.b    = 1;

	/* グローバル変数の場合はこちらで動作する */
	/* Outer and Inner Write-Through, no Write-Allocate */
//	m_attribute.tex  = 0;
//	m_attribute.c    = 1;
//	m_attribute.b    = 0;

	mmu_map_memory(&m_attribute);

#endif /* TOPPERS_SAFEG_SECURE */
}

/*
 *  Vector Table（chip_support.S）
 */
extern void *vector_table;

/*
 *  Disable L4WD0
 */
static void
l4wd0_disable(void)
{
	/* reset watchdog  */ 
	sil_wrw_mem((uint32_t*)RSTMGR_PERMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_PERMODRST) | RSTMGR_PERMODRST_L4WD0);
	/* clear reset watchdog  */
	sil_wrw_mem((uint32_t*)RSTMGR_PERMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_PERMODRST) & ~RSTMGR_PERMODRST_L4WD0);
}

/*
 *  Enable Bridge
 */
static bool_t
enable_bridges(void)
{
	/* reset bridges  */ 
	sil_wrw_mem((uint32_t*)RSTMGR_BRGMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_BRGMODRST) |
				(RSTMGR_BRGMODRST_H2F|RSTMGR_BRGMODRST_LWH2F|RSTMGR_BRGMODRST_F2H));

	/* ToDo error check */

	/* clear reset bridges  */ 
	sil_wrw_mem((uint32_t*)RSTMGR_BRGMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_BRGMODRST) &
				~(RSTMGR_BRGMODRST_H2F|RSTMGR_BRGMODRST_LWH2F|RSTMGR_BRGMODRST_F2H));

	/* Map hps2fpga and lwhp2fpga */
	sil_wrw_mem((uint32_t*)L3REGS_REMAP,
				sil_rew_mem((uint32_t*)L3REGS_REMAP) | (L3REGS_REMAP_HPS2FPGA|L3REGS_REMAP_LWHP2FPGA));

	return true;
}

#ifdef TOPPERS_SAFEG_SECURE
/*
 *  Fist swith to NT code(target_support.S)
 */
extern uint8_t first_switch_to_nt;
extern uint8_t first_switch_to_nt_end;

static void
copy_first_switch_to_nt(void){
	int cnt;
	volatile uint8_t *dst_addr;

	dst_addr = (uint8_t*)OSRAM_ADDR;
	for (cnt = 0; cnt < (&first_switch_to_nt_end - &first_switch_to_nt); cnt++) {
		*(dst_addr + cnt) = *(volatile uint8_t *)(&first_switch_to_nt + cnt);
	}

	__asm__ volatile("dsb");
	__asm__ volatile("isb");
}
#endif /* TOPPERS_SAFEG_SECURE */

/*
 *  L2キャッシュの初期化
 */
static void l2cache_init(void)
{
	uint32_t aux_ctrl = 0;

	aux_ctrl |= (1 << L2X0_AUX_CTRL_SHARE_OVERRIDE_SHIFT) |
	  (1 << L2X0_AUX_CTRL_DATA_PREFETCH_SHIFT) |
		(1 << L2X0_AUX_CTRL_INSTR_PREFETCH_SHIFT);

	pl310_init(aux_ctrl, ~0x0U);
}

#ifdef TOPPERS_SAFEG_SECURE
#include "syscalls_api.h"

uint32_t
safeg_l2cache_init(const uint32_t core_id,
				   const uint32_t ns,
				   const uint32_t a,
				   const uint32_t b,
				   const uint32_t c)
{
	l2cache_init();
	return SAFEG_OK(SAFEG_SYSCALL_RET__DONT_SWITCH);
}

struct safeg_syscall l2init_call = {
	.is_t_callable  = 1,
	.is_nt_callable = 1,
	.name = "l2init",
	.function = safeg_l2cache_init,
};

/*
 *  L2キャッシュデバッグレジスタ書き込み
 */
uint32_t
l2cache_debug_set(const uint32_t core_id,
				  const uint32_t ns,
				  const uint32_t a,
				  const uint32_t b,
				  const uint32_t c)
{
	pl310_debug_set(a);
	return SAFEG_OK(SAFEG_SYSCALL_RET__DONT_SWITCH);
}

struct safeg_syscall l2debug_set_call = {
	.is_t_callable  = 1,
	.is_nt_callable = 1,
	.name = "l2debug",
	.function = l2cache_debug_set,
};

#endif /* TOPPERS_SAFEG_SECURE */

void
target_initialize(void)
{

	/* Disbale high vector */
	set_high_vector(false);

	/* Set Vector Table */
	CP15_SET_VBAR((uint32_t) &vector_table);

#ifndef TOPPERS_SAFEG_NONSECURE
	if (x_sense_mprc()) {
		/* Disable Watchdog(L4WD0) */
		l4wd0_disable();
		/* Enable FPGA bridge */
		enable_bridges();
	}
#endif /*  TOPPERS_SAFEG_NONSECURE */

#ifdef TOPPERS_SAFEG_SECURE
	/* Set NULL(0x00) for Linux second core boot */
	sil_wrw_mem((uint32_t*)SYSMGR_CPU1STARTADDR, 0x00);
	copy_first_switch_to_nt();
#endif /* TOPPERS_SAFEG_SECURE */

	/*
	 * Initialize chip
	 */
	chip_initialize();

#ifdef TOPPERS_SAFEG_SECURE
	if (x_sense_mprc()) {
		/*
		 * L2キャッシュ初期化関数をSafeGに登録
		 */
		uint32_t id;
		safeg_syscall_regdyn(&l2init_call, &id);

		/*
		 * L2キャッシュデバッグレジスタ書き込み関数をSafeGに登録
		 */
		safeg_syscall_regdyn(&l2debug_set_call, &id);
	}
#endif /*  TOPPERS_SAFEG_SECURE */

	/*
	 *  Initialize the UART. If we are using a
	 *  global logger (shared by all cores) only
	 *  the master processor (mprc) initializes it.
	 */
#ifdef G_SYSLOG
	if (x_sense_mprc()) {
	}
#else
	
#endif /* G_SYSLOG */

#if !defined(TOPPERS_SAFEG_SECURE) & !defined(TOPPERS_SAFEG_NONSECURE)
	/* Enable L2 cache in case of without safeg */
	if (x_sense_mprc()) {
		l2cache_init();
	}
#endif /* !defined(TOPPERS_SAFEG_SECURE) & !defined(TOPPERS_SAFEG_NONSECURE) */
}

/*
 *  Target-dependent exit processing
 */
void
target_exit(void)
{
	/*
	 *  Chip exit processing
	 */
	chip_exit();

	/*
	 *  Endless loop
	 */
	while(1);
}
